%=========================================================
% Matlab code for CVPR 2006 paper
% Copyright: Amit Agrawal, 2006
% http://www.umiacs.umd.edu/~aagrawal/
% Permitted for personal use and research purpose only
% Refer to the following citation:
%   1.  A. Agrawal, R. Raskar and R. Chellappa, "Edge Suppression by Gradient Field Transformation Using Cross-Projection Tensors",
%   IEEE Conference on Computer Vision and Pattern Recognition (CVPR) 2006
%   2.  A. Agrawal,  R. Raskar,  Shree K. Nayar &  Y. Li,  "Removing Photography Artifacts using Gradient Projection and
%       Flash-Exposure Sampling",  ACM Transactions on Graphics (Proceedings of  SIGGRAPH) 2005
%=========================================================



% Code for removing shadows from a color image
% Inputs : Flash/no-flash image pair. Ambient (no-flash image has shadows)
% Output:
%       Removed shadows from Ambient image
%       Recovered Illumination Map

clear all;close all;clc;


% Threshold for deciding the homogeneous regions
THRESHOLD_SMALL = 1;


% Read image A (Ambient Image)
imgstr = 'Mannequin/MannequinAmbient.png';
disp(sprintf('Reading image %s',imgstr))
Aimage = imread(imgstr);
Aimage = double(Aimage);

% Read image B (Flash Image)
imgstr = 'Mannequin/MannequinFlash.png';
disp(sprintf('Reading image %s',imgstr))
Bimage = imread(imgstr);
Bimage = double(Bimage);

% Convert to YUV color space
Aimage = RGB2YUV(Aimage);
Bimage = RGB2YUV(Bimage);

% zero padding for 2D integration
PAD = 100
Aimage = padarray(Aimage,[PAD PAD],0,'both');
Bimage = padarray(Bimage,[PAD PAD],0,'both');
[H,W,CH] = size(Aimage);


tic


disp('Finding cross diffusion tensor')

sigma = 0.4 %
ss = floor(6*sigma);
if(ss<=3)
    ss = 3;
end
% K_sigma in the paper; gaussian kernel for smoothing
ww = fspecial('gaussian',ss,sigma);



% find G_sigma for image A and image B
% T11, T12 and T22 are elements of matrix G_sigma
[T11,T12,T22,EigD_1,EigD_2,X1,X2,Y1,Y2] = TensorAnalysis(Bimage,ww);
[T11_2,T12_2,T22_2,EigD_1_2,EigD_2_2,X1_2,X2_2,Y1_2,Y2_2] = TensorAnalysis(Aimage,ww);
clear T11_2 T12_2 T22_2 X1_2 X2_2 Y1_2 Y2_2



% L1 = mu2 , L2 = mu1 of paper
% initially set to 1,1 to retain all edges in image A
L1 = ones(H,W);
L2 = ones(H,W);


% if there is an edge in Bimage, set mu1 = 0. mu2 = 1 to remove that edge
% from image A
idx = find(EigD_2 > THRESHOLD_SMALL);
L2(idx) = 0;


% if both A and B are homogeneous
idx = find(EigD_2 < THRESHOLD_SMALL & EigD_2_2 < THRESHOLD_SMALL);
L1(idx) = 0;
L2(idx) = 0;


% Get cross diffusion tensor terms
D11 = L1.*(X1.^2) + L2.*(Y1.^2);
D12 = L1.*(X1.*X2) + L2.*(Y1.*Y2);
D22 = L1.*(X2.^2) + L2.*(Y2.^2);


% allocate some memory
Ap_color = zeros(H,W,3); % Illumination map
App_color = zeros(H,W,3); % shadow free image


% do for each channel
for channel = 1:CH

    disp('=======================================')
    disp(sprintf('Processing channel = %d',channel))

    u = Aimage(:,:,channel);
    Ap = u;


    % find gradient field
    [gx,gy] = CalculateGradients(Ap,0);

    % Affine transformation using tensors
    gx1 = (D11.*gx + D12.*gy);
    gy1 = (D12.*gx + D22.*gy);

    % 2D Integration
    Ap = Integration2D(gx1,gy1,zeros(H,W));

    % 2D Integration of residual gradient field
    App = Integration2D(gx-gx1,gy-gy1,zeros(H,W));

    % save
    Ap_color(:,:,channel) = Ap;
    App_color(:,:,channel) = App;

end
clear Ap App gx gy gx1 gy1


% remove zero padding
Aimage = Aimage(PAD+1:end-PAD,PAD+1:end-PAD,:);
Bimage = Bimage(PAD+1:end-PAD,PAD+1:end-PAD,:);
D11 = D11(PAD+1:end-PAD,PAD+1:end-PAD,:);
D12 = D12(PAD+1:end-PAD,PAD+1:end-PAD,:);
D22 = D22(PAD+1:end-PAD,PAD+1:end-PAD,:);
Ap_color = Ap_color(PAD+1:end-PAD,PAD+1:end-PAD,:);
App_color = App_color(PAD+1:end-PAD,PAD+1:end-PAD,:);


% YUV to RGB transformation
Aimage = YUV2RGB(Aimage);
Bimage = YUV2RGB(Bimage);
Ap_color = YUV2RGB(Ap_color);
App_color = YUV2RGB(App_color);


% for display
Ap_color = Ap_color - min(Ap_color(:));


close all;
myimshow(Aimage);title('Image A')
myimshow(Bimage);title('Image B');
myimshow(D11);title('D11 of cross diffusion tensor')
myimshow(D12);title('D12 of cross diffusion tensor')
myimshow(D22);title('D22 of cross diffusion tensor')
myimshow(App_color);title('Shadow free image');
myimshow(Ap_color);title('Illumination map');

ttime = toc;
disp('Processing Done...')
disp(sprintf('Total time taken = %f',ttime))



